# Copyright 2021 Timo Röhling <roehling@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import argparse
import os
from pathlib import Path
import re
import subprocess
import shlex
import sys

from ..common import find_packages, sort_packages, get_build_var

destdirs = set()


def prepare_argparse():
    p = argparse.ArgumentParser()
    p.add_argument(
        "--sourcedir",
        default=Path("."),
        type=Path,
        help="root directory of the source package",
    )
    p.add_argument(
        "--max-depth",
        default=2,
        type=int,
        help="limit maximum depth when searching for packages recursively",
    )
    p.add_argument(
        "--unlimited-depth",
        action="store_const",
        dest="max_depth",
        const=-1,
        help="do not limit maximum depth when searching for packages",
    )
    p.add_argument(
        "--builddir", default=Path(".rosbuild"), type=Path, help="build directory"
    )
    p.add_argument(
        "--destdir",
        default=Path("debian/tmp"),
        type=Path,
        help="installation directory",
    )
    g = p.add_mutually_exclusive_group()
    g.add_argument(
        "--detect", action="store_true", help="detect if ROS packages are to be built"
    )
    g.add_argument(
        "--build-types", action="store_true", help="list detected build types"
    )
    g.add_argument(
        "--build-order", action="store_true", help="list packages in build order"
    )
    return p


def run():
    p = prepare_argparse()
    args = p.parse_args()

    packages = find_packages(args.sourcedir, max_depth=args.max_depth)

    if args.detect:
        return 0 if packages else 1

    if args.build_types:
        sys.stdout.write(
            "\n".join(sorted(set(p.get_build_type() for _, p in packages)))
        )
        return 0

    packages = sort_packages(packages)
    skipped_packages = get_build_var(None, "ROS_SKIP_PACKAGES")
    if skipped_packages:
        skipped_packages = set(shlex.split(skipped_packages))

    if args.build_order:
        sys.stdout.write("\n".join(p.name for _, p in packages))

    sys.stdout.write(
        "\n"
        "#############################################################################\n"
        "## Detected ROS packages (in build order):\n"
        "##\n"
    )
    for path, package in packages:
        sys.stdout.write(f"## - {package.name} [{package.get_build_type()}]\n")
    sys.stdout.write(
        "#############################################################################\n"
        "\n"
    )

    try:
        global destdirs
        destdirs = set()
        for path, package in packages:
            if package.name in skipped_packages:
                sys.stdout.write(
                    "\n"
                    "#############################################################################\n"
                    f"## Skipping package {package.name}\n"
                    "#############################################################################\n"
                )
                continue

            sys.stdout.write(
                "\n"
                "#############################################################################\n"
                f"## Configuring package {package.name}\n"
                "#############################################################################\n"
            )
            do_configure(args, path, package)
            sys.stdout.write(
                "\n"
                "#############################################################################\n"
                f"## Building package {package.name}\n"
                "#############################################################################\n"
            )
            do_build(args, path, package)
            sys.stdout.write(
                "\n"
                "#############################################################################\n"
                f"## Testing package {package.name}\n"
                "#############################################################################\n"
            )
            do_test(args, path, package)
            sys.stdout.write(
                "\n"
                "#############################################################################\n"
                f"## Installing package {package.name}\n"
                "#############################################################################\n"
            )
            do_install(args, path, package)
        sys.stdout.write(
            "\n"
            "#############################################################################\n"
            "## All ROS packages have been built successfully\n"
            "#############################################################################\n"
        )
    except subprocess.CalledProcessError as e:
        sys.stderr.write(f"ERROR: {str(e)}\n")
        return e.returncode or 1
    return 0


def make_environment(args, path, package):
    build_type = package.get_build_type()
    new_environ = os.environ.copy()
    if build_type in ["ament_python"]:
        package_name = re.sub(r"[^A-Za-z0-9]", "_", package.name)
        pybuild_package_keys = [
            key
            for key in os.environ
            if key.startswith("PYBUILD_") and key.endswith(f"_{package_name}")
        ]
        for key in pybuild_package_keys:
            new_environ[key[: -len(package_name) + 1]] = os.environ[key]
        if "PYBUILD_DESTDIR" not in new_environ:
            new_environ["PYBUILD_DESTDIR"] = args.destdir
    return new_environ


def make_default_args(args, path, package):
    build_type = package.get_build_type()
    args = [f"--sourcedir={path}", f"--builddir={args.builddir / package.name}"]
    if build_type in ["ament_python"]:
        args += ["--buildsystem=pybuild"]
    elif build_type in ["ament_cmake", "catkin", "cmake"]:
        args += ["--buildsystem=cmake"]
    return args


def do_configure(args, path, package):
    global destdirs
    cmdline = ["dh_auto_configure"] + make_default_args(args, path, package) + ["--"]
    environ = make_environment(args, path, package)
    environ["CMAKE_PREFIX_PATH"] = ":".join(str(d / "usr") for d in destdirs)

    build_type = package.get_build_type()
    if build_type == "catkin":
        cmdline += ["-DCATKIN_BUILD_BINARY_PACKAGE=ON"]

    if build_type in ["ament_cmake", "catkin", "cmake"]:
        extra_args = get_build_var(package.name, "ROS_CONFIGURE_ARGS")
        if extra_args:
            cmdline += shlex.split(extra_args)

    sys.stdout.write(shlex.join(cmdline))
    sys.stdout.write("\n")
    sys.stdout.flush()
    subprocess.check_call(cmdline, env=environ)


def do_build(args, path, package):
    cmdline = ["dh_auto_build"] + make_default_args(args, path, package)
    environ = make_environment(args, path, package)
    extra_args = get_build_var(package.name, "ROS_BUILD_ARGS")
    if extra_args:
        cmdline += ["--"] + shlex.split(extra_args)

    sys.stdout.write(shlex.join(cmdline))
    sys.stdout.write("\n")
    sys.stdout.flush()
    subprocess.check_call(cmdline, env=environ)


def do_test(args, path, package):
    buildopts = re.split(r"[ ,\t\r\n]+", os.environ.get("DEB_BUILD_OPTIONS", ""))
    if "nocheck" in buildopts:
        sys.stdout.write("skipping tests because of 'nocheck' build option\n")
        return
    cmdline = ["dh_auto_test"] + make_default_args(args, path, package)
    environ = make_environment(args, path, package)
    extra_args = get_build_var(package.name, "ROS_TEST_ARGS")
    if extra_args:
        cmdline += ["--"] + shlex.split(extra_args)

    sys.stdout.write(shlex.join(cmdline))
    sys.stdout.write("\n")
    sys.stdout.flush()
    subprocess.check_call(cmdline, env=environ)


def do_install(args, path, package):
    global destdirs
    destdir = Path(get_build_var(package.name, "ROS_DESTDIR", args.destdir))
    destdirs.add(destdir)
    cmdline = (
        ["dh_auto_install"]
        + make_default_args(args, path, package)
        + [f"--destdir={destdir}"]
    )
    environ = make_environment(args, path, package)
    extra_args = get_build_var(package.name, "ROS_INSTALL_ARGS")
    if extra_args:
        cmdline += ["--"] + shlex.split(extra_args)

    sys.stdout.write(shlex.join(cmdline))
    sys.stdout.write("\n")
    sys.stdout.flush()
    subprocess.check_call(cmdline, env=environ)
